use anyhow::Result;
use std::convert::TryFrom;
use thiserror::Error;

static INPUT: &str = include_str!("../../input05");

fn main() -> Result<()> {
    part1()?;
    part2()?;
    Ok(())

}

fn part1() -> Result<()> {
    let reader = BinaryReader::from(INPUT);
    let max_seat_id: Result<u32, Errors> = reader.into_iter().try_fold(0, |max, seat| {
        let seat_id = seat?.calculate_sead_id();
        if seat_id > max {
            Ok(seat_id)
        } else {
            Ok(max)
        }
    });
    println!("max seat id: {}", max_seat_id?);
    Ok(())
}

fn part2() -> Result<()> {
    let mut seats = BinaryReader::from(INPUT)
        .into_iter()
        .filter_map(|seat| match seat {
            Ok(seat) => Some(seat.calculate_sead_id()),
            Err(_) => None,
        })
        .collect::<Vec<u32>>();
    seats.sort();
    let mut last = None;
    for seat in seats.iter() {
        if let Some(l) = last {
            if l != seat - 1 {
                println!("my seat is: {}", seat - 1);
                break
            };
        };
        last = Some(*seat)
    }
    Ok(())
}

#[derive(Error, Debug)]
enum Errors<'a> {
    #[error("invalid input ({0}): {1}")]
    InvalidInput(&'a str, &'a str),
}

struct BinaryReader<'a> {
    inner: &'a str,
}

impl<'a> From<&'a str> for BinaryReader<'a> {
    fn from(input: &'a str) -> Self {
        BinaryReader {
            inner: input.trim(),
        }
    }
}

impl<'a> Iterator for BinaryReader<'a> {
    type Item = Result<Seat, Errors<'a>>;

    fn next(&mut self) -> Option<Self::Item> {
        let mut splitted = self.inner.splitn(2, '\n');
        let line = splitted.next().unwrap();
        self.inner = splitted.next().or(Some(""))?;

        match line {
            "" => None,
            line => Some(Seat::try_from(line)),
        }
    }
}

#[derive(Debug, Eq, PartialEq)]
struct Seat {
    row: u32,
    col: u32,
}

impl<'a> TryFrom<&'a str> for Seat {
    type Error = Errors<'a>;

    fn try_from(data: &'a str) -> Result<Self, Self::Error> {
        if data.len() != 10 {
            return Err(Errors::InvalidInput("seat code must be 10 chars log", data));
        }
        let row = u32_from_str(&data[0..7], 'B', 'F')?;
        let col = u32_from_str(&data[7..10], 'R', 'L')?;
        Ok(Seat { row, col })
    }
}

impl Seat {
    fn calculate_sead_id(&self) -> u32 {
        self.row * 8 + self.col
    }
}

fn u32_from_str<'a>(input: &'a str, one: char, zero: char) -> Result<u32, Errors<'a>> {
    let mut res = 0;
    for c in input.chars() {
        res = if c == zero {
            res << 1
        } else if c == one {
            (res << 1) + 1
        } else {
            return Err(Errors::InvalidInput("invalid char", input));
        }
    }

    Ok(res)
}

#[cfg(test)]
mod test {
    use super::*;
    #[test]
    fn examples_part1() {
        let seat1 = Seat::try_from("BFFFBBFRRR").unwrap();
        assert_eq!(seat1, Seat { row: 70, col: 7 });
        assert_eq!(seat1.calculate_sead_id(), 567);
        let seat2 = Seat::try_from("FFFBBBFRRR").unwrap();
        assert_eq!(seat2, Seat { row: 14, col: 7 });
        assert_eq!(seat2.calculate_sead_id(), 119);
        let seat3 = Seat::try_from("BBFFBBFRLL").unwrap();
        assert_eq!(seat3, Seat { row: 102, col: 4 });
        assert_eq!(seat3.calculate_sead_id(), 820);
    }

    #[test]
    fn failures() {
        let reader = BinaryReader::from(INPUT);
        let max_seat_id: Result<u32, Errors> = reader.into_iter().try_fold(0, |max, seat| {
            let seat_id = seat?.calculate_sead_id();
            if seat_id > max {
                Ok(seat_id)
            } else {
                Ok(max)
            }
        });
        assert!(max_seat_id.unwrap() > 861);
    }
}
